{$A+,B-,C-,D+,E-,F-,G+,H+,I-,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y-,Z1}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00100000}
{$IMAGEBASE $00400000}
{$APPTYPE GUI}
unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, Spin, StdCtrls, ExtCtrls, Buttons, WinSock, ShellAPI;

const
  WSA_msg = $1244;

  StartupErrorStr : string = ' .'#13#10' :'#13#10'%s';
  ServerErrorStr  : string = '  .'#13#10' :'#13#10'%s';
  ClientErrorStr  : string = '    .'#13#10' :'#13#10'%s';
  DNSErrorStr     : string = '  IP-  .'#13#10' :'#13#10'%s';
type
  TMainForm = class(TForm)
    Memo1: TMemo;
    Panel1: TPanel;
    Edit1: TEdit;
    Label1: TLabel;
    SpinEdit1: TSpinEdit;
    Label2: TLabel;
    Panel2: TPanel;
    Memo2: TMemo;
    StatusBar1: TStatusBar;
    Panel3: TPanel;
    Panel4: TPanel;
    ClientBtn: TSpeedButton;
    ServerBtn: TSpeedButton;
    Bevel1: TBevel;
    Bevel2: TBevel;
    SendBtn: TButton;
    CheckBox2: TCheckBox;
    Label3: TLabel;
    procedure Panel2MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormResize(Sender: TObject);
    procedure SendBtnClick(Sender: TObject);
    procedure ClientBtnClick(Sender: TObject);
    procedure ServerBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Memo1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Panel1Resize(Sender: TObject);
    procedure Label3Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure WSAHandle( var msg : TMessage); message WSA_msg;
  end;

var
  MainForm  : TMainForm;
  WSAHandle : HWND = 0;
  Server    : TSocket = INVALID_SOCKET;
  Client    : TSocket = INVALID_SOCKET;

  procedure NetInit( Handle : HWND);
  procedure NetDone;
  procedure OpenServer( Port : Integer );
  procedure OpenClient(const Address : string; Port:Integer );
  procedure CloseConnection;

implementation

{$R *.DFM}

function WSAErrorMessage( Code : Integer ) : string;
Begin
  case Code of
   WSANOTINITIALISED: Result:='      WSAStartup().   .';
   WSAENETDOWN: Result:='  Windows    .';
   WSAEINPROGRESS: Result:='  Windows Sockets  .   .';
   WSAEMFILE: Result:='      .  ,    ,   .';
   WSAEPROTONOSUPPORT: Result:=' TCP/IP  UDP      Windows.';
   WSAEPROTOTYPE:Result:='      .   .';
   WSAESOCKTNOSUPPORT:Result:='        .   .';
   WSAEADDRINUSE :Result:='   .     -   .';
   WSAEFAULT:Result:=' namelen  .   .';
   WSAEAFNOSUPPORT:Result:='       .   .';
   WSAEINVAL:Result:='         .   .';
   WSAENOBUFS:Result:='  .   ,    ,   .';
   WSAENOTSOCK:Result:='    .   .';
   WSAEISCONN:Result:='     .   .';
   WSAEOPNOTSUPP:Result:='   ,    listen.   .';
   WSAENOTCONN:Result:='     .   .';
   10049:Result:='  .';
   else
     Result:='  ('+IntToStr(Code)+').';
 end;
end;


procedure TMainForm.Panel2MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if ssLeft in Shift then
  begin
    Memo1.Height:=Memo1.Height+Y-Panel2.Height div 2;
  end;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  Bevel1.Width:=Panel1.ClientWidth-4;
end;

function DetectServerIP(const Host:string ) : Integer;
var
  HostEnt : PHostEnt;
begin
  result:=inet_addr(PChar(Host));
  if result = -1 then
  begin
    HostEnt:=gethostbyname(PChar(Host));
    if HostEnt=nil then
      raise Exception.CreateFmt(DNSErrorStr,[SysErrorMessage( WSAGetLastError )]);
    result:=Integer(Pointer(HostEnt.h_addr^)^);
  end;
end;


procedure NetInit(Handle : HWND );
var
  WSAdata : TWSAData;
  Error   : Integer;
begin
  WSAHandle:=Handle;
  Error:=WSAStartup($101,WSAdata);
  if Error <> 0 then
  begin
    raise Exception.CreateFmt(StartupErrorStr,[WSAErrorMessage( WSAGetLastError )]);
  end;
  Client:=INVALID_SOCKET;
  Server:=INVALID_SOCKET;
end;

procedure NetDone;
begin
  CloseConnection;
  WSACleanup;
end;


procedure OpenServer(Port:Integer);
var
  Addr : TSockAddr;

  procedure CheckError( Error : Integer );
  var
    Code : Integer;
  begin
    if Error = SOCKET_ERROR then
    begin
      Code:=WSAGetLastError;
      if Server<>INVALID_SOCKET then
      begin
        CloseSocket(Server);
        Server:=INVALID_SOCKET;
      end;
      raise Exception.CreateFmt(ServerErrorStr,[WSAErrorMessage( Code )]);
    end;
  end;

begin
  //  
  Server:=Socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
  if Server = INVALID_SOCKET then CheckError(SOCKET_ERROR);

  //  
  Addr.sin_family:=AF_INET;
  Addr.sin_port:=htons(Port);
  Addr.sin_addr.S_addr:=0;
  CheckError( bind( Server, Addr, Sizeof(TSockAddr)) );

  //     
  CheckError(WSAAsyncSelect( Server, WSAHandle, WSA_Msg, FD_READ or FD_ACCEPT or FD_CLOSE));

  //     
  CheckError( listen(Server,100) );
end;


procedure OpenClient(const Address:string; Port : Integer);
var
  Addr : TSockAddr;
  procedure CheckError( Error : Integer );
  var
    Code : Integer;
  begin
    if Error = SOCKET_ERROR then
    begin
      Code:=WSAGetLastError;
      if Code = WSAEWOULDBLOCK then exit;
      if Client<>INVALID_SOCKET then
      begin
        CloseSocket(Client);
        Client:=INVALID_SOCKET;
      end;
      raise Exception.CreateFmt(ClientErrorStr,[WSAErrorMessage(Code)]);
    end;
  end;
begin
  //  
  Addr.sin_family:=AF_INET;
  Addr.sin_port:=htons(Port);
  Addr.sin_addr.S_addr:=DetectServerIP(Address);

  //  
  Client:=Socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
  if Client = INVALID_SOCKET then CheckError(SOCKET_ERROR);

  //    
  CheckError(WSAAsyncSelect(Client,WSAHandle,WSA_MSG,FD_CONNECT or FD_Write or FD_READ or FD_CLOSE or FD_WRITE));

  CheckError(Connect(Client,Addr,Sizeof(TSockAddr)));
end;

Procedure CloseConnection;
begin
  if Client<>INVALID_SOCKET then
  begin
    CloseSocket(Client);
    Client:=INVALID_SOCKET;
  end;
  if Server<>INVALID_SOCKET then
  begin
    CloseSocket(Server);
    Server:=INVALID_SOCKET;
  end;
end;

procedure TMainForm.WSAHandle;
const
  BufSize = $400;
var
  Count : integer;
  i     : integer;
  buf   : array[0..BufSize] of char;
  ws    : string;
begin
  case msg.LParam of
    FD_CONNECT:
      begin
        StatusBar1.SimpleText:=' .';
        SendBtn.Enabled:=True;
      end;
    FD_CLOSE:
      begin
        if ClientBtn.Down then ClientBtn.Down:=False;
        SendBtn.Enabled:=False;
        StatusBar1.SimpleText:=' .';
      end;
    FD_ACCEPT:
      begin
        if Client<>-1 then CloseSocket(Client);
        Client:=Accept(Server,nil,nil);
        StatusBar1.SimpleText:='    .';
        SendBtn.Enabled:=True;
      end;
    FD_READ:
      while true do
      begin
        Count:=Recv(Client,buf,BufSize,0);
        if Count<=0 then break;
        ws:='';
        for i:=0 to Count-1 do
        begin
          case ord(buf[i]) of
            0..9,11,12,14..31,ord('#'):
              ws:=ws+'#'+IntToHex(ord(buf[i]),2)
            else
              ws:=ws+buf[i];
          end;
        end;
        Memo2.SelStart:=Length(Memo2.Text);
        Memo2.SelText:=ws;
        Memo2.SelStart:=Length(Memo2.Text);
      end;
  end;
end;

procedure TMainForm.SendBtnClick(Sender: TObject);
var
  i      : integer;
  Src    : string;
  Dst    : string;

function HexCharToInt(c:Char):integer;
begin
  case c of
    '0'..'9':result:=ord(c)-ord('0');
    'A'..'F':result:=ord(c)-ord('A')+10;
    'a'..'f':result:=ord(c)-ord('a')+10;
    else
      raise ERangeError.Create('Error char');
  end;
end;

begin
  if not SendBtn.Enabled then exit;


  if CheckBox2.Checked then
  begin
    Src:=Memo1.Text;
    Dst:='';
    i:=1;
    while i<=length(Src) do
    begin
      if (Src[i]='#') then
      begin
        inc(i);
        if i>length(Src) then
        begin
          Dst:=Dst + '#';
          inc(i);
        end
        else
        if i<=length(Src)-1 then
        begin
          Dst:=Dst + chr(HexCharToInt(Src[i]) shl 4 + HexCharToInt(Src[i+1]));
          inc(i,2);
        end;
      end
      else
      if (Src[i]='\') and (i < Length(Src)) then
      begin
        case Src[i+1] of
          'n':Dst:=Dst + #10;
          'r':Dst:=Dst + #13;
          't':Dst:=Dst + #9;
          'b':Dst:=Dst + #7;
          '0':Dst:=Dst + #0;
          else
          begin
            Dst:=Dst+Src[i];
            dec(i);
          end;
          inc(i,2);
        end;
      end
      else
      begin
        Dst:=Dst + Src[i];
        inc(i);
      end;
    end;
  end
  else
    Dst:=Memo1.Text;

  Send(Client,Dst[1],Length(Dst),0);
end;

procedure TMainForm.ClientBtnClick(Sender: TObject);
begin
  CloseConnection;
  StatusBar1.SimpleText:='';
  if ClientBtn.Down then
  begin
    try
      OpenClient(Edit1.text, SpinEdit1.Value);
      StatusBar1.SimpleText:=' .  ...';
    except
      ClientBtn.Down:=False;
      raise;
    end;
  end;
end;

procedure TMainForm.ServerBtnClick(Sender: TObject);
begin
  CloseConnection;
  StatusBar1.SimpleText:='';
  if ServerBtn.Down then
  begin
    try
      OpenServer(SpinEdit1.Value);
      StatusBar1.SimpleText:=' .  ...';
    except
      ServerBtn.Down:=False;
      raise;
    end;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  NetInit(Handle);
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  NetDone;
end;

procedure TMainForm.Memo1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (ssCtrl in Shift) and (Key=13) then
  begin
    SendBtnClick(nil);
    Key:=0;
  end;
end;


procedure TMainForm.Panel1Resize(Sender: TObject);
begin
  Label3.Left:=Panel1.Width - Label3.Width-8;
end;

procedure TMainForm.Label3Click(Sender: TObject);
begin
  ShellExecute(0,nil,'http://www.dynsoft.ru',nil,nil,SW_SHOW);
end;

end.
